class Student:
    def __init__(self, first_name: str, last_name: str, email: str):
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

    def __eq__(self, other) -> bool:
        return (self.first_name == other.first_name
                and self.last_name == other.last_name
                and self.email == other.email)

    def to_tuple(self) -> tuple[str, str, str]:
        """
        Get a tuple of strings representation of a student.
        """
        return self.first_name, self.last_name, self.email



import json
import logging
from importlib import resources
from pathlib import Path
import jsonschema

from . import config, schemas, sheets
from .students import Student
from .teams import Team

SUBMISSION_INFO_FILE_NAME = "submission.json"
FEEDBACK_DIR_NAME = "feedback"
FEEDBACK_COLLECTED_DIR_NAME = "feedback_collected"


class Submission:
    def __init__(self, team_dir: Path):
        self.root_dir = team_dir
        self.sheet = sheets.Sheet(self.root_dir.parent)
        self._load()

    def get_feedback_dir(self) -> Path:
        return self.root_dir / FEEDBACK_DIR_NAME

    def get_collected_feedback_dir(self) -> Path:
        return self.root_dir / FEEDBACK_COLLECTED_DIR_NAME

    def get_collected_feedback_path(self) -> Path:
        """
        Return the collected feedback file. This can be
        either a single pdf file, or a single zip archive. Throw an error if neither
        exists.
        """
        collected_feedback_dir = self.get_collected_feedback_dir()
        if not collected_feedback_dir.is_dir():
            logging.critical(
                "The directory for collected feedback at"
                f" '{str(collected_feedback_dir)}' does not exist. You probably"
                " have to run the 'collect' command first."
            )
        collected_feedback_files = list(collected_feedback_dir.iterdir())
        assert (
                len(collected_feedback_files) == 1
                and collected_feedback_files[0].is_file()
                and collected_feedback_files[0].suffix in [".pdf", ".zip"]
        )
        return collected_feedback_files[0]

    def get_combined_feedback_file(self) -> Path:
        """
        Return the combined feedback file. This is always a
        zip archive because in the usual case it will contain feedback from multiple
        tutors.
        """
        combined_feedback_dir = self.sheet.get_combined_feedback_path()
        if not combined_feedback_dir.is_dir():
            logging.critical(
                "The directory for combined feedback at"
                f" '{str(combined_feedback_dir)}' does not exist. You probably"
                " have to run the 'combine' command first."
            )
        combined_feedback_files = list(
            (combined_feedback_dir / self.root_dir.name).iterdir()
        )
        assert (
                len(combined_feedback_files) == 1
                and combined_feedback_files[0].is_file()
                and combined_feedback_files[0].suffix == ".zip"
        )
        return combined_feedback_files[0]

    def _load(self):
        """
        Load the submission info of the team directory. In particular the
        team, the ADAM ID and if it is a relevant team.
        """
        try:
            with open(
                    self.root_dir / SUBMISSION_INFO_FILE_NAME, "r", encoding="utf-8"
            ) as submission_info_file:
                submission_info = json.load(submission_info_file)
                submission_info_schema = json.loads(
                    resources.files(schemas).joinpath("submission-info-schema.json").read_text(encoding="utf-8"))
                jsonschema.validate(submission_info, submission_info_schema, jsonschema.Draft7Validator)
                self.team = Team(
                    [Student(*student) for student
                     in submission_info.get("team")],
                    submission_info.get("adam_id")
                )
                self.relevant = submission_info.get("relevant")
        except FileNotFoundError:
            logging.critical("The submission.json file does not exist.")
        except NotADirectoryError:
            logging.critical(f"The path '{self.root_dir}' is not a team directory.")
        except jsonschema.exceptions.ValidationError as error:
            logging.critical(f"The submission.json file does not have the right format: {error.message}")

    def __lt__(self, other):
        return self.root_dir < other.root_dir


def create_submission_info_file(_the_config: config.Config, team: Team,
                                team_id: str, team_dir: Path) -> None:
    """
    Write in each team directory a JSON file which contains the team,
    the ADAM ID of the team which ADAM sets anew with each exercise sheet,
    and if the tutor specified in the config has to mark this team.
    """
    team = team.to_tuples()
    relevant_teams = [team.to_tuples() for team in
                      _the_config.get_relevant_teams()]
    is_relevant = False
    submission_info = {}
    if team in relevant_teams:
        is_relevant = True
    submission_info.update({"team": team})
    submission_info.update({"adam_id": team_id})
    submission_info.update({"relevant": is_relevant})
    with open(
            team_dir / SUBMISSION_INFO_FILE_NAME, "w", encoding="utf-8"
    ) as submission_info_file:
        json.dump(
            submission_info,
            submission_info_file,
            indent=4,
            ensure_ascii=False,
        )



import logging


def unsupported_marking_mode_error(marking_mode) -> None:
    """
    Throw an error if a marking mode is encountered that is not handled
    correctly. This is primarily used in the else case of if-elif-else branches
    on the marking_mode, which shouldn't be reached in regular operation anyway
    because the config settings should be validated first. But this function
    offers an easy way to find sections to consider if we ever want to add a new
    marking_mode.
    """
    logging.critical(f"Unsupported marking mode {marking_mode}!")



import json
import logging
from importlib import resources
from pathlib import Path
import jsonschema

from . import errors, schemas
from .teams import *


# Within this class, Team objects are created with their adam_id set to None
# because the adam_id is not available at the time of the Config class
# instantiation.
class Config:
    def __init__(self, config_paths: list[Path]) -> None:
        data = {}
        for path in config_paths:
            logging.info(f"Reading config file '{path}'")
            data.update(json.loads(path.read_text(encoding="utf-8")))
        config_schema = json.loads(resources.files(schemas).joinpath("config-schema.json").read_text(encoding="utf-8"))
        jsonschema.validate(data, config_schema, jsonschema.Draft7Validator)

        for key, value in data.items():
            setattr(self, key, value)

        # We currently plan to support the following marking modes.
        # static:   Every tutor corrects the submissions of the teams assigned to
        #           that tutor. These will usually be the teams in that tutors
        #           exercise class.
        # exercise: Every tutor corrects some exercise(s) on all sheets.
        if self.marking_mode == "static":
            if self.tutor_name not in self.classes:
                logging.critical(f"Did not find a class for '{self.tutor_name}' in the config.")
            self.teams = [team for classs in self.classes.values() for team in classs]
        else:
            if self.tutor_name not in self.tutor_list:
                logging.critical(f"Did not find '{self.tutor_name}' in tutor_list in the config.")

        # Sort teams and their students to make iterating over them
        # predictable, independent of their order in config.json.
        for team in self.teams:
            team.sort()
        self.teams.sort()

        # Create Team objects with their adam_id set to None because
        # the adam_id is not available here
        if self.marking_mode == "static":
            self.classes = {
                tutor: [Team([Student(*student) for student in team], None)
                        for team in teams]
                for tutor, teams in self.classes.items()
            }
        self.teams = [Team([Student(*student) for student in team], None)
                      for team in self.teams]
        _validate_teams(self.teams, self.max_team_size)
        logging.info("Processed config successfully.")

    def get_relevant_teams(self) -> list[Team]:
        """
        Get a list of teams that the tutor specified in the config has to mark.
        We rename the directories using the `DO_NOT_MARK_PREFIX` and thereafter only
        access relevant teams via `get_relevant_submissions()`.
        """
        if self.marking_mode == "static":
            return self.classes[self.tutor_name]
        elif self.marking_mode == "exercise":
            return self.teams
        else:
            errors.unsupported_marking_mode_error(self.marking_mode)
            return []


def _validate_teams(teams: list[Team], max_team_size) -> None:
    """
    Check for duplicate entries and maximal team size.
    """
    all_students: list[tuple[str, str]] = []
    all_emails: list[str] = []
    for team in teams:
        if len(team.members) > max_team_size:
            logging.critical(f"Team with size {len(team.members)} violates maximal "
                             f"team size.")
        for member in team.members:
            all_students.append((member.first_name, member.last_name))
            all_emails.append(member.email)
    if len(all_students) != len(set(all_students)):
        logging.critical("There are duplicate students in the config file!")
    if len(all_emails) != len(set(all_emails)):
        logging.critical("There are duplicate student emails in the config file!")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import pathlib
import pytest
import shutil
import subprocess
import typing

CONFIG_INDIVIDUAL = pathlib.Path("config-individual.json")
CONFIG_STATIC = pathlib.Path("config-shared-static.json")
CONFIG_EXERCISE = pathlib.Path("config-shared-exercise.json")
SAMPLE_SHEET = pathlib.Path("Sample Sheet.zip")
SAMPLE_SHEET_DIR = pathlib.Path("Sample Sheet")
SAMPLE_SHEET_SUB_DIR = SAMPLE_SHEET_DIR / "SUB_DIR"


@pytest.fixture(autouse=True)
def change_test_dir(request, monkeypatch):
    """
    Make the directory where the test module lies the current working directory.
    """
    monkeypatch.chdir(request.fspath.dirname)


@pytest.fixture(params=["Abgaben", "Submissions"])
def setup_test_directory(request, tmp_path: pathlib.Path):
    """
    Copy sample sheet directory, create zip file in ADAM's format, and remove
    sheet directory.
    """
    shutil.copytree(SAMPLE_SHEET_DIR, tmp_path / SAMPLE_SHEET_DIR.name)
    shutil.move(
        tmp_path / SAMPLE_SHEET_SUB_DIR,
        tmp_path / SAMPLE_SHEET_DIR / request.param,
    )
    os.makedirs(tmp_path / "temp")
    shutil.move(tmp_path / SAMPLE_SHEET_DIR, tmp_path / "temp")
    shutil.move(tmp_path / "temp", tmp_path / SAMPLE_SHEET_DIR)
    shutil.make_archive(
        str(tmp_path / SAMPLE_SHEET_DIR), "zip", tmp_path / SAMPLE_SHEET_DIR
    )
    shutil.rmtree(tmp_path / SAMPLE_SHEET_DIR)
    # Copy individual config file.
    shutil.copy(CONFIG_INDIVIDUAL, tmp_path)
    return tmp_path


@pytest.fixture
def config_shared(request, monkeypatch, setup_test_directory: pathlib.Path):
    """
    This fixture copies all relevant files into a testing directory.
    """
    shutil.copy(request.param, setup_test_directory)
    monkeypatch.chdir(setup_test_directory)
    return request.param


@pytest.fixture(params=["tamara", "tatiana", "terence"])
def insert_tutor_name(request):
    with open("config-individual.json", "r") as f:
        filled_in = f.read().replace("PLACEHOLDER_NAME", request.param)
    with open("config-individual.json", "w") as f:
        f.write(filled_in)


@pytest.fixture(params=["true", "false"])
def insert_xopp_setting(request):
    with open("config-individual.json", "r") as f:
        filled_in = f.read().replace("PLACEHOLDER_XOPP_SETTING", request.param)
    with open("config-individual.json", "w") as f:
        f.write(filled_in)


def give_feedback():
    # Remove '.todo' suffixes.
    for todo_file in pathlib.Path.cwd().glob("**/*.todo"):
        shutil.move(todo_file, todo_file.with_suffix(""))
    # Enter points.
    for point_file in pathlib.Path.cwd().glob("**/points*.json"):
        with open(point_file, "r") as file:
            data = file.read()
        data = data.replace(': ""', ': "1.5"')
        with open(point_file, "w") as file:
            file.write(data)


@pytest.mark.parametrize(
    "config_shared, args",
    [
        (CONFIG_STATIC, []),
        (CONFIG_EXERCISE, ["-e", "1", "3"]),
    ],
    indirect=["config_shared"],
)
def test(
    capfd,
    config_shared: pathlib.Path,
    insert_tutor_name,
    insert_xopp_setting,
    args: list[str],
):
    # Call 'init'.
    subprocess.check_call(
        [
            "krummstab",
            "-i",
            str(CONFIG_INDIVIDUAL),
            "-s",
            str(config_shared),
            "init",
            str(SAMPLE_SHEET),
        ]
        + args
    )
    # Remove zip.
    SAMPLE_SHEET.unlink()

    # Verify 'init' ran successfully.
    out, err = capfd.readouterr()
    assert "Command 'init' terminated successfully." in out

    # Prepare for 'collect'.
    give_feedback()

    # Call 'collect'.
    subprocess.check_call(
        [
            "krummstab",
            "-i",
            str(CONFIG_INDIVIDUAL),
            "-s",
            str(config_shared),
            "collect",
            str(SAMPLE_SHEET_DIR),
        ]
    )

    # Verify 'collect' ran successfully.
    out, err = capfd.readouterr()
    assert "Command 'collect' terminated successfully." in out

    # Call 'combine' for the configurations using the mode 'exercise'.
    if config_shared == CONFIG_EXERCISE:
        subprocess.check_call(
            [
                "krummstab",
                "-i",
                str(CONFIG_INDIVIDUAL),
                "-s",
                str(config_shared),
                "combine",
                str(SAMPLE_SHEET_DIR),
            ]
        )

        # Verify 'combine' ran successfully.
        out, err = capfd.readouterr()
        assert "Command 'combine' terminated successfully." in out

    # Call 'send'.
    subprocess.check_call(
        [
            "krummstab",
            "-i",
            str(CONFIG_INDIVIDUAL),
            "-s",
            str(config_shared),
            "send",
            "--dry_run",
            str(SAMPLE_SHEET_DIR),
        ]
    )
    # Verify 'send' ran successfully.
    out, err = capfd.readouterr()
    assert "Command 'send' terminated successfully." in out



import json
import logging
from collections import defaultdict
from pathlib import Path
import xlsxwriter
from xlsxwriter import Workbook
from xlsxwriter.utility import xl_rowcol_to_cell, xl_range_abs
from xlsxwriter.worksheet import Worksheet

from .. import config
from ..teams import *

BOLD = {'bold': True}
BORDER = {'border': 1}
BORDER_LEFT = {'left': 1}
BORDER_TOP_BOTTOM = {'top': 1, 'bottom': 1}
GRAY = {'bg_color': '#E2E2E2'}
GREEN = {'bg_color': '#9BE189'}
YELLOW = {'bg_color': '#FFE699'}
RED = {'bg_color': '#EE7868'}
PLAGIARISM_RED = {'bg_color': 'red'}
PERCENT = {'num_format': '0%'}
TEXT_WRAP = {'text_wrap': True}

PERCENTAGE_TO_PASS_CELL = '$B$4'
IMPROVE_AVG_RED_CELL = '$E$3'
IMPROVE_AVG_GREEN_CELL = '$E$1'

PLAGIARISM = "Plagiarism"


def add_legend(workbook: Workbook, worksheet: Worksheet):
    """
    Adds a legend that explains the conditional formatting.
    """
    worksheet.set_row(0, 28)
    worksheet.set_row(1, 28)
    worksheet.set_row(2, 28)
    worksheet.set_row(3, 28)
    worksheet.merge_range(
        0, 0, 0, 2, "Pass", workbook.add_format(GREEN | BORDER)
    )
    worksheet.merge_range(
        1, 0, 1, 2, "Fail", workbook.add_format(RED | BORDER)
    )
    worksheet.merge_range(
        2, 0, 2, 1, "2 x Plagiarism",
        workbook.add_format(PLAGIARISM_RED | BORDER)
    )
    worksheet.write(
        0, 3,
        "Does not need to improve average\nif percentage is lower "
        "than:",
        workbook.add_format(GREEN | TEXT_WRAP)
    )
    worksheet.write_number(
        0, 4,
        -0.05, workbook.add_format(GREEN | PERCENT)
    )
    worksheet.write(
        1, 3,
        "Should improve average\nif percentage is between:",
        workbook.add_format(YELLOW | TEXT_WRAP)
    )
    worksheet.write_formula(
        1, 4,
        f'TEXT({IMPROVE_AVG_GREEN_CELL},"0%")'
        f'& " to " & TEXT({IMPROVE_AVG_RED_CELL},"0%")',
        workbook.add_format(YELLOW | PERCENT)
    )
    worksheet.write(
        2, 3,
        "Has to improve average\nby at least the following percentage:",
        workbook.add_format(RED | TEXT_WRAP)
    )
    worksheet.write_number(
        2, 4,
        0.10, workbook.add_format(RED | PERCENT)
    )
    worksheet.write(
        3, 0, "Percentage of points\nrequired to pass:",
        workbook.add_format(TEXT_WRAP)
    )
    worksheet.write_number(
        3, 1, 0.5, workbook.add_format(PERCENT)
    )


def add_pass_or_fail_conditional_formatting(workbook: Workbook, worksheet: Worksheet, row: int,
                                            graded_sheet_names, all_sheet_names):
    """
    Colors the name red if not enough points can be collected.
    Colors the name green when enough points have been collected.
    """
    num_legend_rows = 4
    total_points_all_sheets_range = xl_range_abs(
        1 + num_legend_rows, 5, 1 + num_legend_rows, 5 + len(all_sheet_names) - 1
    )
    possible_points_range = xl_range_abs(
        1 + num_legend_rows, 5 + len(graded_sheet_names), 1 + num_legend_rows, 5 + len(all_sheet_names) - 1
    )
    worksheet.conditional_format(row, 0, row, 2, {
        'type': 'formula',
        'criteria': f"=(SUM(INDIRECT(ADDRESS(ROW(),3)), SUM({possible_points_range})))"
                    f" < (SUM({total_points_all_sheets_range}) * {PERCENTAGE_TO_PASS_CELL})",
        'format': workbook.add_format(RED)
    })
    worksheet.conditional_format(row, 0, row, 2, {
        'type': 'formula',
        'criteria': f"INDIRECT(ADDRESS(ROW(),3)) >= (SUM("
                    f"{total_points_all_sheets_range}) * {PERCENTAGE_TO_PASS_CELL})",
        'format': workbook.add_format(GREEN)
    })


def add_plagiarism_conditional_formatting(workbook: Workbook, worksheet: Worksheet,
                                          row: int, all_sheet_names):
    """
    Colors the name bright red if there are two plagiarisms.
    """
    student_marks_range = xl_range_abs(row, 5, row, 5 + len(all_sheet_names) - 1)
    worksheet.conditional_format(row, 0, row, 1, {
        'type': 'formula',
        'criteria': f'=COUNTIF({student_marks_range},"{PLAGIARISM}") >= 2',
        'format': workbook.add_format(PLAGIARISM_RED)
    })


def add_average_conditional_formatting(workbook: Workbook, worksheet: Worksheet, row: int):
    """
    Adds the conditional formatting for the current average and the percentage
    by which the average must be improved.
    """
    worksheet.conditional_format(row, 3, row, 4, {
        'type': 'formula',
        'criteria': f'=INDIRECT(ADDRESS(ROW(),5)) <= {IMPROVE_AVG_GREEN_CELL}',
        'format': workbook.add_format(GREEN)
    })
    worksheet.conditional_format(row, 3, row, 4, {
        'type': 'formula',
        'criteria': f'=INDIRECT(ADDRESS(ROW(),5)) <= {IMPROVE_AVG_RED_CELL}',
        'format': workbook.add_format(YELLOW)
    })
    worksheet.conditional_format(row, 3, row, 4, {
        'type': 'formula',
        'criteria': f'=INDIRECT(ADDRESS(ROW(),5)) > {IMPROVE_AVG_RED_CELL} ',
        'format': workbook.add_format(RED)
    })


def add_student_average(workbook: Workbook, worksheet: Worksheet, row: int, graded_sheet_names, all_sheet_names):
    """
    Calculates a student's current weighted average and the percentage
    by which the average must be improved to pass with the remaining
    exercise sheets.
    """
    num_legend_rows = 4
    student_marks_range = xl_range_abs(row, 5, row, 5 + len(all_sheet_names) - 1)
    total_points_all_sheets_range = xl_range_abs(
        1 + num_legend_rows, 5, 1 + num_legend_rows, 5 + len(all_sheet_names) - 1
    )
    possible_points_range = xl_range_abs(
        1 + num_legend_rows, 5 + len(graded_sheet_names), 1 + num_legend_rows, 5 + len(all_sheet_names) - 1
    )
    worksheet.write_formula(row, 3,
                            f'=IFERROR(SUMPRODUCT(ISNUMBER({student_marks_range})*1,{student_marks_range},'
                            f' {total_points_all_sheets_range}) / SUMPRODUCT(ISNUMBER({student_marks_range})*1,'
                            f' {total_points_all_sheets_range}),"")',
                            workbook.add_format(BORDER))
    worksheet.write_formula(row, 4,
                            f'=IFERROR((SUM('
                            f'{total_points_all_sheets_range})*{PERCENTAGE_TO_PASS_CELL}'
                            f' - (INDIRECT(ADDRESS(ROW(),3))))/(COLUMNS({possible_points_range})'
                            f'*(INDIRECT(ADDRESS(ROW(),4))))-1,"")',
                            workbook.add_format(BORDER | PERCENT))


def write_mark(worksheet, row, col, mark) -> None:
    """
    Writes a mark in the cell specified by row and column.
    """
    if mark is None:
        worksheet.write_blank(row, col, None)
    elif isinstance(mark, str) and not mark.replace(".", "", 1).isdigit():
        worksheet.write(row, col, mark)
    else:
        worksheet.write_number(row, col, float(mark))


def add_student_marks_worksheet_points_per_sheet(workbook: Workbook, worksheet: Worksheet,
                                                 _the_config: config.Config, row, email,
                                                 students_marks, all_sheet_names, graded_sheet_names,
                                                 points_per_sheet_cell_addresses):
    """
    Writes the points per sheet of a student in the worksheet for the sheets and calculates
    the sum to get the total points of the student.
    If points are given per exercise: Uses cell addresses from the points per
    exercise worksheet that contain the calculated points per sheet.
    """
    if _the_config.points_per == 'exercise':
        for col, cell_address in enumerate(points_per_sheet_cell_addresses[email], start=5):
            worksheet.write(row, col, f"={cell_address}")
    else:
        for col, sheet_name in enumerate(list(all_sheet_names)[:len(graded_sheet_names)], start=5):
            mark = students_marks[email].get(sheet_name)
            write_mark(worksheet, row, col, mark)
    student_marks_range = xl_range_abs(row, 5, row, 5 + len(all_sheet_names) - 1)
    worksheet.write_formula(row, 2, f"=SUM({student_marks_range})", workbook.add_format(BORDER))


def add_student_marks_worksheet_points_per_exercise(workbook: Workbook, worksheet: Worksheet,
                                                    row, email, students_marks, all_sheet_names,
                                                    graded_sheet_names) -> list[str]:
    """
    Writes the points per exercise of a student in the worksheet for the exercises, calculates
    the sum to get the points per sheet and handles plagiarism. Returns a list of
    cell addresses containing the calculated points per sheet.
    """
    col = 2
    points_per_sheet_cell_addresses = []
    for sheet_name in list(all_sheet_names)[:len(graded_sheet_names)]:
        student_marks_range = xl_range_abs(row, col + 1, row, col + len(graded_sheet_names[sheet_name]))
        worksheet.write_dynamic_array_formula(
            row, col, row, col,
            f'=IF(COUNTIF({student_marks_range},"{PLAGIARISM}") > 0,"{PLAGIARISM}",'
            f'IF(AND(NOT(ISNUMBER({student_marks_range}))),"",SUM({student_marks_range})))',
            workbook.add_format(BORDER_LEFT)
        )
        cell_address = xl_rowcol_to_cell(row, col)
        points_per_sheet_cell_addresses.append(f"'Points Per Exercise'!{cell_address}")
        col += 1
        for exercise in sorted(graded_sheet_names[sheet_name]):
            mark = students_marks[email][sheet_name].get(exercise)
            write_mark(worksheet, row, col, mark)
            col += 1
    return points_per_sheet_cell_addresses


def add_student_name(workbook, worksheet, row, name) -> None:
    """
    Writes the specified name of the student in the first two columns
    for the specified row. Changes the background color of the row
    alternately to gray for better readability.
    """
    first_name, last_name = name
    if row % 2 == 0:
        worksheet.write(row, 0, first_name, workbook.add_format(BORDER_TOP_BOTTOM | GRAY))
        worksheet.write(row, 1, last_name, workbook.add_format(BORDER_TOP_BOTTOM | GRAY))
        worksheet.set_row(row, None, workbook.add_format(GRAY))
    else:
        worksheet.write(row, 0, first_name, workbook.add_format(BORDER_TOP_BOTTOM))
        worksheet.write(row, 1, last_name, workbook.add_format(BORDER_TOP_BOTTOM))


def create_worksheet_points_per_exercise(workbook: Workbook, email_to_name, students_marks, all_sheet_names,
                                         graded_sheet_names) -> defaultdict[str, list[str]]:
    """
    Creates an Excel worksheet that contains the points per exercise. Returns a dict with the
    students' email addresses and the cell addresses that contain the calculated points per sheet.
    """
    worksheet = workbook.add_worksheet("Points Per Exercise")
    worksheet.write(0, 0, "First Name", workbook.add_format(BOLD))
    worksheet.write(0, 1, "Last Name", workbook.add_format(BOLD))
    col = 2
    for sheet_name in all_sheet_names:
        worksheet.write(0, col, sheet_name, workbook.add_format(BOLD | BORDER_LEFT))
        if sheet_name in graded_sheet_names:
            col += 1
            for exercise in sorted(graded_sheet_names[sheet_name]):
                worksheet.write(0, col, exercise.replace("exercise", "task"), workbook.add_format(BOLD))
                col += 1
        else:
            col += 1
    all_points_per_sheet_cell_addresses = defaultdict(list)
    student_start_row = 1
    sorted_emails = sorted(email_to_name.keys(), key=lambda e: (email_to_name[e][0], email_to_name[e][1]))
    for row, email in enumerate(sorted_emails, start=student_start_row):
        add_student_name(workbook, worksheet, row, email_to_name[email])
        points_per_sheet_cell_addresses = add_student_marks_worksheet_points_per_exercise(
            workbook, worksheet, row, email, students_marks, all_sheet_names, graded_sheet_names
        )
        all_points_per_sheet_cell_addresses[email] = points_per_sheet_cell_addresses
    worksheet.autofit()
    return all_points_per_sheet_cell_addresses


def create_worksheet_points_per_sheet(workbook: Workbook, _the_config: config.Config,
                                      email_to_name, students_marks,
                                      all_sheet_names, graded_sheet_names,
                                      points_per_sheet_cell_addresses) -> None:
    """
    Creates an Excel worksheet that contains the points per sheet,
    useful values calculated with functions, and conditional formatting.
    """
    worksheet = workbook.add_worksheet("Points Summary")
    worksheet.activate()
    num_legend_rows = 4
    worksheet.write(3 + num_legend_rows, 0, "First Name", workbook.add_format(BOLD))
    worksheet.write(3 + num_legend_rows, 1, "Last Name", workbook.add_format(BOLD))
    worksheet.write(3 + num_legend_rows, 2, "Total Points", workbook.add_format(BOLD))
    worksheet.write(3 + num_legend_rows, 3, "Current Average", workbook.add_format(BOLD))
    worksheet.write(3 + num_legend_rows, 4, "Improve", workbook.add_format(BOLD))
    worksheet.write(1 + num_legend_rows, 0, "Max Points", workbook.add_format(BOLD))
    worksheet.write(2 + num_legend_rows, 0, "Average", workbook.add_format(BOLD))

    worksheet.set_row(0 + num_legend_rows, cell_format=workbook.add_format(BORDER_TOP_BOTTOM))
    student_start_row = 4 + num_legend_rows
    student_end_row = student_start_row + len(email_to_name) - 1
    for col, sheet_name in enumerate(all_sheet_names, start=5):
        worksheet.write(0 + num_legend_rows, col, sheet_name, workbook.add_format(BOLD | BORDER_TOP_BOTTOM))
        max_points_value = _the_config.max_points_per_sheet.get(sheet_name)
        worksheet.write(1 + num_legend_rows, col, max_points_value)
        avg_range = xl_range_abs(student_start_row, col, student_end_row, col)
        worksheet.write_formula(2 + num_legend_rows, col, f'=IFERROR(AVERAGE({avg_range}),"")')

    sorted_emails = sorted(email_to_name.keys(), key=lambda e: (email_to_name[e][0], email_to_name[e][1]))
    for row, email in enumerate(sorted_emails, start=student_start_row):
        add_student_name(workbook, worksheet, row, email_to_name[email])
        add_student_marks_worksheet_points_per_sheet(workbook, worksheet, _the_config,
                                                     row, email, students_marks,
                                                     all_sheet_names, graded_sheet_names,
                                                     points_per_sheet_cell_addresses)
        add_student_average(workbook, worksheet, row, graded_sheet_names, all_sheet_names)
        add_average_conditional_formatting(workbook, worksheet, row)
        add_plagiarism_conditional_formatting(workbook, worksheet, row, all_sheet_names)
        add_pass_or_fail_conditional_formatting(workbook, worksheet, row, graded_sheet_names, all_sheet_names)
    add_legend(workbook, worksheet)
    worksheet.autofit()


def load_marks_files(marks_dir: Path, _the_config: config.Config):
    """
    Loads the data of the individual marks files in the specified directory.
    """
    marks_files = marks_dir.glob("points_*_individual.json")
    if _the_config.points_per == 'exercise':
        students_marks = defaultdict(lambda: defaultdict(dict))
        graded_sheet_names = defaultdict(list)
    else:
        students_marks = defaultdict(dict)
        graded_sheet_names = set()
    tutors = defaultdict(set)
    for file in marks_files:
        with open(file, 'r') as f:
            data = json.load(f)
            sheet_name = data["adam_sheet_name"]
            marks = data["marks"]
            tutors[sheet_name].add(data["tutor_name"])
            if _the_config.points_per == 'exercise':
                for email, exercises in marks.items():
                    for exercise, mark in exercises.items():
                        if students_marks[email][sheet_name].get(exercise):
                            logging.warning(f'{exercise} of sheet {sheet_name} is marked multiple times for {email}!')
                        students_marks[email][sheet_name][exercise] = mark
                        if exercise not in graded_sheet_names[sheet_name]:
                            graded_sheet_names[sheet_name].append(exercise)
            else:
                graded_sheet_names.add(sheet_name)
                for email, mark in marks.items():
                    if students_marks[email].get(sheet_name):
                        logging.warning(f'Sheet {sheet_name} is marked multiple times for {email}!')
                    students_marks[email][sheet_name] = mark
    for sheet_name, tutor_list in tutors.items():
        for tutor in _the_config.classes if _the_config.marking_mode == 'static' else _the_config.tutor_list:
            if tutor not in tutor_list:
                logging.warning(f'There is no file from tutor {tutor} for sheet {sheet_name}!')
    return students_marks, graded_sheet_names


def create_marks_summary_excel_file(_the_config: config.Config, marks_dir: Path) -> None:
    """
    Generates an Excel file that summarizes the students' marks. Uses a path
    to a directory containing the individual marks files.
    """
    email_to_name = create_email_to_name_dict(_the_config.teams)
    workbook = xlsxwriter.Workbook("Points_Summary_Report.xlsx")
    students_marks, graded_sheet_names = load_marks_files(marks_dir, _the_config)
    all_sheet_names = _the_config.max_points_per_sheet.keys()
    points_per_sheet_cell_addresses = None
    if _the_config.points_per == 'exercise':
        points_per_sheet_cell_addresses = create_worksheet_points_per_exercise(
            workbook, email_to_name, students_marks, all_sheet_names, graded_sheet_names
        )
    create_worksheet_points_per_sheet(
        workbook, _the_config, email_to_name, students_marks, all_sheet_names, graded_sheet_names,
        points_per_sheet_cell_addresses
    )
    workbook.close()


def summarize(_the_config: config.Config, args) -> None:
    """
    Generate an Excel file summarizing students' marks after the individual
    marks files have been collected in a directory.
    """
    if not args.marks_dir.is_dir():
        logging.critical("The given individual marks directory is not valid!")
    create_marks_summary_excel_file(
        _the_config, args.marks_dir
    )



import argparse
import pathlib

from .commands import *


DEFAULT_SHARED_CONFIG_FILE = "config-shared.json"
DEFAULT_INDIVIDUAL_CONFIG_FILE = "config-individual.json"


def add_parsers():
    parser = argparse.ArgumentParser(description="")
    add_main_command_parser(parser)
    subparsers = add_subcommand_parser(parser)
    add_help_command_parser(subparsers)
    add_init_command_parser(subparsers)
    add_collect_command_parser(subparsers)
    add_combine_command_parser(subparsers)
    add_send_command_parser(subparsers)
    add_summarize_command_parser(subparsers)
    return parser


def add_main_command_parser(parser):
    parser.add_argument(
        "-s",
        "--config-shared",
        default=DEFAULT_SHARED_CONFIG_FILE,
        type=pathlib.Path,
        help=(
            "path to the json config file containing shared settings such as "
            "the student/email list"
        ),
    )
    parser.add_argument(
        "-i",
        "--config-individual",
        default=DEFAULT_INDIVIDUAL_CONFIG_FILE,
        type=pathlib.Path,
        help=(
            "path to the json config file containing individual settings such "
            "as tutor name and email configuration"
        ),
    )


def add_subcommand_parser(parser):
    subparsers = parser.add_subparsers(
        required=True,
        help="available sub-commands",
        dest="sub_command",
    )
    return subparsers


def add_help_command_parser(subparsers):
    parser_init = subparsers.add_parser(
        "help",
        help=(
            "print this help message; "
            "run e.g. 'krummstab init -h' to print the help of a sub-command"
        ),
    )


def add_init_command_parser(subparsers):
    parser_init = subparsers.add_parser(
        "init",
        help="unpack zip file from ADAM and prepare directory structure",
    )
    parser_init.add_argument(
        "adam_zip_path",
        type=pathlib.Path,
        help="path to the zip file downloaded from ADAM",
    )
    parser_init.add_argument(
        "-t",
        "--target",
        type=pathlib.Path,
        required=False,
        help="path to the directory that will contain the submissions",
    )
    group = parser_init.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "-n",
        "--num-exercises",
        dest="num_exercises",
        type=int,
        help="the number of exercises in the sheet",
    )
    group.add_argument(
        "-e",
        "--exercises",
        dest="exercises",
        nargs="+",
        type=int,
        help="the exercises you have to mark",
    )
    parser_init.set_defaults(func=init)


def add_collect_command_parser(subparsers):
    parser_collect = subparsers.add_parser(
        "collect",
        help="collect feedback files after marking is done",
    )
    parser_collect.add_argument(
        "sheet_root_dir",
        type=pathlib.Path,
        help="path to the sheet's directory",
    )
    parser_collect.set_defaults(func=collect)


def add_combine_command_parser(subparsers):
    parser_combine = subparsers.add_parser(
        "combine",
        help=(
            "combine multiple share archives, only necessary if tutors mark per"
            " exercise and have to integrate their individual feedback into a"
            " single ZIP file to send to the students"
        ),
    )
    parser_combine.add_argument(
        "sheet_root_dir",
        type=pathlib.Path,
        help="path to the sheet's directory",
    )
    parser_combine.set_defaults(func=combine)


def add_send_command_parser(subparsers):
    parser_send = subparsers.add_parser(
        "send",
        help="send feedback via email",
    )
    parser_send.add_argument(
        "sheet_root_dir",
        type=pathlib.Path,
        help="path to the sheet's directory",
    )
    parser_send.add_argument(
        "-d",
        "--dry_run",
        action="store_true",
        help="only print emails instead of sending them",
    )
    parser_send.set_defaults(func=send)


def add_summarize_command_parser(subparsers):
    parser_summarize = subparsers.add_parser(
        "summarize",
        help="summarize individual marks files into Excel report",
    )
    parser_summarize.add_argument(
        "marks_dir",
        type=pathlib.Path,
        help="path to the directory with all individual marks files",
    )
    parser_summarize.set_defaults(func=summarize)



from collections.abc import Iterator
from pathlib import Path
import logging
import json

from . import config, submissions, errors

SHEET_INFO_FILE_NAME = "sheet.json"
FEEDBACK_FILE_PREFIX = "feedback_"
COMBINED_DIR_NAME = "feedback_combined"
DO_NOT_MARK_PREFIX = "DO_NOT_MARK_"
SHARE_ARCHIVE_PREFIX = "share_archive"


class Sheet:
    def __init__(self, sheet_root_dir: Path):
        self._sheet_info_path = sheet_root_dir / SHEET_INFO_FILE_NAME
        if not sheet_root_dir.is_dir() or not self._sheet_info_path.exists():
            logging.critical("Could not find a directory with marking "
                             f"information at '{sheet_root_dir}'. Use the command "
                             "'init' to set it up.")
        self.root_dir = sheet_root_dir
        self._load()

    def _load(self):
        with open(self._sheet_info_path, "r", encoding="utf-8") as sheet_info_file:
            sheet_info = json.load(sheet_info_file)
            self.name = sheet_info.get('adam_sheet_name')
            self.exercises = sheet_info.get('exercises')

    def get_adam_sheet_name_string(self) -> str:
        """
        Turn the sheet name given by ADAM into a string usable for file names.
        """
        return self.name.replace(" ", "_").lower()

    def get_feedback_file_name(self, _the_config: config.Config) -> str:
        file_name = FEEDBACK_FILE_PREFIX + self.get_adam_sheet_name_string() + "_"
        if _the_config.marking_mode == "exercise":
            file_name += _the_config.tutor_name + "_"
            file_name += "_".join([f"ex{exercise}" for exercise in self.exercises])
        elif _the_config.marking_mode == "static":
            # Remove trailing underscore.
            file_name = file_name[:-1]
        else:
            errors.unsupported_marking_mode_error(_the_config.marking_mode)
        return file_name

    def get_combined_feedback_file_name(self) -> str:
        return FEEDBACK_FILE_PREFIX + self.get_adam_sheet_name_string()

    def get_combined_feedback_path(self) -> Path:
        return self.root_dir / COMBINED_DIR_NAME

    def get_marks_file_path(self, _the_config: config.Config) -> Path:
        return (
                self.root_dir
                / f"points_{_the_config.tutor_name.lower()}_{self.get_adam_sheet_name_string()}.json"
        )

    def get_individual_marks_file_path(self, _the_config: config.Config) -> Path:
        marks_file_path = self.get_marks_file_path(_the_config)
        individual_marks_file_path = marks_file_path.with_name(
            marks_file_path.stem + "_individual" + marks_file_path.suffix
        )
        return individual_marks_file_path

    def get_all_team_submission_info(self) -> Iterator[submissions.Submission]:
        """
        Return all team submission info. Exclude other directories that may be created
        in the sheet root directory, such as one containing combined feedback.
        """
        for sub_dir in self.root_dir.iterdir():
            if sub_dir.is_dir() and sub_dir != self.get_combined_feedback_path():
                yield submissions.Submission(sub_dir)

    def get_relevant_submissions(self) -> Iterator[submissions.Submission]:
        """
        Return the submission info of the teams whose submission has to be
        corrected by the tutor running the script.
        """
        for submission in self.get_all_team_submission_info():
            if submission.relevant:
                yield submission

    def get_share_archive_file_path(self) -> Path:
        return self.root_dir / (
                SHARE_ARCHIVE_PREFIX
                + f"_{self.get_adam_sheet_name_string()}_"
                + "_".join([f"ex{num}" for num in self.exercises])
                + ".zip"
        )

    def get_share_archive_files(self) -> Iterator[Path]:
        """
        Return all share archive files under the current sheet root dir.
        """
        for share_archive_file in self.root_dir.glob(
                SHARE_ARCHIVE_PREFIX + "*.zip"
        ):
            yield share_archive_file


def create_sheet_info_file(sheet_root_dir: Path, adam_sheet_name: str,
                           _the_config: config.Config, exercises=None) -> Sheet:
    """
    Write information generated during the execution of the `init` command in a
    sheet info file. In particular the name of the exercise sheet as given by
    ADAM and which exercises should be marked if the marking mode is `exercise`.
    Later commands (e.g. `collect`, or `send`) are meant to load the information
    stored in this file when initializing the class sheet_info.Sheet
    and access it that way.
    """
    info_dict = {}
    info_dict["adam_sheet_name"] = adam_sheet_name
    if _the_config.marking_mode == "exercise":
        info_dict["exercises"] = exercises
    with open(
            sheet_root_dir / SHEET_INFO_FILE_NAME, "w", encoding="utf-8"
    ) as sheet_info_file:
        json.dump(
            info_dict,
            sheet_info_file,
            indent=4,
            ensure_ascii=False,
            sort_keys=True,
        )
    return Sheet(sheet_root_dir=sheet_root_dir)



import logging
import mimetypes
import os
import pathlib
import smtplib
import textwrap

from email.message import EmailMessage
from getpass import getpass

from .. import config, errors, sheets, utils


def add_attachment(mail: EmailMessage, path: pathlib.Path) -> None:
    """
    Add a file as attachment to an email.
    This is copied from Patrick's/Silvan's script, not entirely sure what's
    going on here.
    """
    assert path.exists()
    ctype, encoding = mimetypes.guess_type(path)
    if ctype is None or encoding is not None:
        ctype = "application/octet-stream"
    maintype, subtype = ctype.split("/", 1)
    with open(path, "rb") as fp:
        mail.add_attachment(
            fp.read(),
            maintype=maintype,
            subtype=subtype,
            filename=os.path.basename(path),
        )


def construct_email(
    receivers: list[str],
    cc: list[str],
    subject: str,
    content: str,
    sender: str,
    attachment: pathlib.Path,
) -> EmailMessage:
    """
    Construct an email object.
    """
    assert isinstance(receivers, list)
    assert subject and content and sender and attachment
    mail = EmailMessage()
    mail.set_content(content)
    mail["Subject"] = subject
    mail["From"] = sender
    mail["To"] = ", ".join(receivers)
    if cc:
        mail["Cc"] = ", ".join(cc)
    add_attachment(mail, attachment)
    return mail


def email_to_text(email: EmailMessage) -> None:
    to = email["To"]
    cc = email["CC"]
    subject = email["Subject"]
    content = ""
    attachments = []
    for part in email.walk():
        if part.is_attachment():
            attachments.append(part.get_filename())
        elif not part.is_multipart():
            content += part.get_content()
    lines = []

    def format_line(left: str, right: str) -> str:
        return f"\033[0;34m{left}\033[0m {right}"

    lines.append(format_line("To:", to))
    if cc:
        lines.append(format_line("CC:", cc))
    if attachments:
        lines.append(format_line("Attachments:", ", ".join(attachments)))
    lines.append(format_line("Subject:", subject))
    lines.append(format_line("Start Body", ""))
    if content[-1] == "\n":
        content = content[:-1]
    lines.append(content)
    lines.append(format_line("End Body", ""))
    return "\n".join(lines)


def print_emails(emails: list[EmailMessage]) -> None:
    separator = "\n\033[0;33m" + 80 * "=" + "\033[0m\n"
    email_strings = [email_to_text(email) for email in emails]
    print(f"{separator}{separator.join(email_strings)}{separator}")


def send_messages(emails: list[EmailMessage], _the_config: config.Config) -> None:
    with smtplib.SMTP(_the_config.smtp_url, _the_config.smtp_port) as smtp:
        smtp.starttls()
        if _the_config.smtp_user:
            logging.warning(
                "The setting 'smtp_user' should probably be empty for the"
                " 'send' command to work, trying anyway."
            )
            password = getpass("Email password: ")
            smtp.login(_the_config.smtp_user, password)
        for email in emails:
            logging.info(f"Sending email to {email['To']}")
            # During testing, I didn't manage to trigger the exceptions below.
            # Additionally `refused_recipients` was always empty, even when the
            # documentation of smtplib states that it should be populated when
            # some but not all of the recipients are refused. Instead I always
            # get receive an email from the Outlook server containing the error
            # message.
            try:
                refused_recipients = smtp.send_message(email)
            except smtplib.SMTPRecipientsRefused:
                logging.warning(
                    f"Email to '{email['To']}' failed to deliver because all"
                    " recipients were refused."
                )
            except smtplib.SMTPSenderRefused:
                logging.critical(
                    "Email sender was refused, failed to deliver any emails."
                )
            except (
                smtplib.SMTPHeloError,
                smtplib.SMTPDataError,
                smtplib.SMTPNotSupportedError,
            ):
                logging.warning(
                    f"Email to '{email['To']}' failed to deliver because of"
                    " some weird error."
                )
            for refused_recipient, (
                smtp_error,
                error_message,
            ) in refused_recipients.items():
                logging.warning(
                    f"Email to '{refused_recipient}' failed to deliver because"
                    " the recipient was refused with the SMTP error code"
                    f" '{smtp_error}' and the message '{error_message}'."
                )
        logging.info("Done sending emails.")


def get_team_email_subject(_the_config: config.Config, sheet: sheets.Sheet) -> str:
    """
    Builds the email subject.
    """
    return f"Feedback {sheet.name} | {_the_config.lecture_title}"


def get_assistant_email_subject(_the_config: config.Config, sheet: sheets.Sheet) -> str:
    """
    Builds the email subject.
    """
    return f"Marks for {sheet.name} | {_the_config.lecture_title}"


def get_email_greeting(name_list: list[str]) -> str:
    """
    Builds the first line of the email.
    """
    # Only keep one name per entry, "Hans Jakob" becomes "Hans".
    name_list = [name.split(" ")[0] for name in name_list]
    name_list.sort()
    assert len(name_list) > 0
    if len(name_list) == 1:
        names = name_list[0]
    elif len(name_list) == 2:
        names = name_list[0] + " and " + name_list[1]
    else:
        names = ", ".join(name_list[:-1]) + ", and " + name_list[-1]
    return "Dear " + names + ","


def get_team_email_content(name_list: list[str], _the_config: config.Config, sheet: sheets.Sheet) -> str:
    """
    Builds the body of the email that sends feedback to students.
    """
    return textwrap.dedent(
        f"""
    {get_email_greeting(name_list)}

    Please find feedback on your submission for {sheet.name} in the attachment.
    If you have any questions, you can contact us in the exercise session or by replying to this email (reply to all).

    Best,
    {_the_config.email_signature}"""  # noqa
    )[
        1:
    ]  # Removes the leading newline.


def get_assistant_email_content(_the_config: config.Config, sheet: sheets.Sheet) -> str:
    """
    Builds the body of the email that sends the points to the assistant.
    """
    return textwrap.dedent(
        f"""
    Dear assistant for {_the_config.lecture_title}

    Please find my marks for {sheet.name} in the attachment.

    Best,
    {_the_config.email_signature}"""
    )[
        1:
    ]  # Removes the leading newline.


def get_assistant_email_attachment_path(_the_config: config.Config, sheet: sheets.Sheet) -> pathlib.Path:
    """
    Sending the marks file to the assistant as is has the disadvantage that
    points are only listed per team. This makes it difficult for the assistant
    to figure out how many points each individual student has. So we use the
    individual marks file where points are listed per student.
    """
    return sheet.get_individual_marks_file_path(_the_config)


def create_email_to_team(submission, _the_config: config.Config, sheet: sheets.Sheet):
    team_first_names = submission.team.get_first_names()
    team_emails = submission.team.get_emails()
    if _the_config.marking_mode == "exercise":
        feedback_file_path = submission.get_combined_feedback_file()
    elif _the_config.marking_mode == "static":
        feedback_file_path = submission.get_collected_feedback_path()
    else:
        errors.unsupported_marking_mode_error(_the_config.marking_mode)
    return construct_email(
        list(team_emails),
        _the_config.feedback_email_cc,
        get_team_email_subject(_the_config, sheet),
        get_team_email_content(team_first_names, _the_config, sheet),
        _the_config.tutor_email,
        feedback_file_path,
    )


def create_email_to_assistant(_the_config: config.Config, sheet: sheets.Sheet):
    return construct_email(
        [_the_config.assistant_email],
        _the_config.feedback_email_cc,
        get_assistant_email_subject(_the_config, sheet),
        get_assistant_email_content(_the_config, sheet),
        _the_config.tutor_email,
        get_assistant_email_attachment_path(_the_config, sheet),
    )


def send(_the_config: config.Config, args) -> None:
    """
    After the collection step finished successfully, send the feedback to the
    students via email. This currently only works if the tutor's email account
    is whitelisted for the smtp-ext.unibas.ch server, or if the tutor uses
    smtp.unibas.ch with an empty smtp_user.
    """
    # Prepare.
    sheet = sheets.Sheet(args.sheet_root_dir)
    # Send emails.
    emails: list[EmailMessage] = []
    for submission in sheet.get_relevant_submissions():
        emails.append(create_email_to_team(submission, _the_config, sheet))
    # TODO: As of now the plan is to only send assistant emails if the marking
    # mode is "static" because there the assistant collects the points
    # centrally. In case of "exercise", we plan to distribute the point files
    # through the share_archives, so there is no need to send an email to the
    # assistant, but this may change in the future.
    if _the_config.marking_mode != "exercise" and _the_config.assistant_email:
        emails.append(create_email_to_assistant(_the_config, sheet))
    logging.info(f"Drafted {len(emails)} email(s).")
    if args.dry_run:
        logging.info("Sending emails now would send the following emails:")
        print_emails(emails)
        logging.info("No emails sent.")
    else:
        print_emails(emails)
        really_send = utils.query_yes_no(
            (
                f"Do you really want to send the {len(emails)} email(s) "
                "printed above?"
            ),
            default=False,
        )
        if really_send:
            send_messages(emails, _the_config)
        else:
            logging.info("No emails sent.")



import json
import logging
import shutil
import subprocess
from typing import Optional
from zipfile import ZipFile

from .. import config, sheets, submissions, utils
from .summarize import PLAGIARISM

def validate_marks_json(_the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Verify that all necessary marks are present in the MARK_FILE_NAME file and
    adhere to the granularity defined in the config file.
    """
    marks_json_file = sheet.get_marks_file_path(_the_config)
    if not marks_json_file.is_file():
        logging.critical(
            f"Missing points file in directory '{sheet.root_dir}'!"
        )
    with open(marks_json_file, "r", encoding="utf-8") as marks_file:
        marks = json.load(marks_file)
    relevant_teams = []
    for submission in sheet.get_relevant_submissions():
        relevant_teams.append(submission.team.get_team_key())
    marked_teams = list(marks.keys())
    if sorted(relevant_teams) != sorted(marked_teams):
        logging.critical(
            "There is no 1-to-1 mapping between teams "
            "that need to be marked and entries in the "
            f"'{marks_json_file.name}' "
            "file! Make sure that it contains exactly one entry for every team "
            "that needs to be marked, and that the team keys "
            "have the correct format. The team key has to match the content of "
            "the submission.json file in the team directory and consist of the "
            "ADAM ID and the alphabetically sorted last names of all team "
            "members in the following format: ID_Last-Name1_Last-Name2"
        )
    if _the_config.points_per == "exercise":
        marks_list = [
            mark
            for team_marks in marks.values()
            for mark in team_marks.values()
        ]
    else:
        marks_list = marks.values()
    if not all(marks_list):
        logging.critical(
            f"There are missing points in the '{marks_json_file.name}' file!"
        )
    if not all(
        (float(mark) / _the_config.min_point_unit).is_integer()
        for mark in marks_list if mark != PLAGIARISM
    ):
        logging.critical(
            f"'{marks_json_file.name}' contains marks that are more"
            " fine-grained than allowed! You may only award points in"
            f" '{_the_config.min_point_unit}' increments."
        )


def collect_feedback_files(submission: submissions.Submission,
                           _the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Take the contents of a {team_dir}/feedback directory and collect the files
    that actually contain feedback (e.g., no .xopp files). If there are
    multiple, add them to a zip archive and save it to
    {team_dir}/feedback_collected. If there is only a single pdf, copy it to
    {team_dir}/feedback_collected.
    """
    feedback_dir = submission.get_feedback_dir()
    collected_feedback_dir = submission.get_collected_feedback_dir()
    collected_feedback_zip_name = sheet.get_feedback_file_name(_the_config) + ".zip"
    # Error handling.
    if not feedback_dir.exists():
        logging.critical(
            f"Missing feedback directory for team {submission.root_dir.name}!"
        )
    content = list(feedback_dir.iterdir())
    if any(".todo" in file_or_dir.name for file_or_dir in content):
        logging.critical(
            f"Feedback for {submission.root_dir.name} contains placeholder TODO file!"
        )
    # The directory for collected feedback should exist and be empty. Either it
    # was created new, or the user chose to overwrite and previously existing
    # directories have been removed and replaced by empty ones.
    assert collected_feedback_dir.is_dir() and not any(
        collected_feedback_dir.iterdir()
    )
    # Create list of feedback files. Those are all files in the feedback
    # directory which do not have an ignored suffix.
    feedback_files = [
        file
        for file in feedback_dir.rglob("*")
        if file.is_file() and file.suffix not in _the_config.ignore_feedback_suffix
    ]
    # Ask for confirmation if the feedback directory contains hidden files that
    # are maybe not supposed to be part of the collected feedback.
    hidden_files = [f for f in feedback_files if utils.is_hidden_file(f.name)]
    for hidden_file in hidden_files:
        include_anyway = utils.query_yes_no(
            (
                "There seem to be hidden files in your feedback directory, "
                f"e.g. '{str(hidden_file)}'. Do you want to include them in "
                "your feedback anyway? (Consider adding ignored suffixes in "
                "your individual configuration file to avoid this prompt in "
                "the future.)"
            ),
            default=False,
        )
        if not include_anyway:
            feedback_files.remove(hidden_file)

    if not feedback_files:
        logging.critical(
            f"Feedback archive for team {submission.root_dir.name} is empty!"
        )

    # If there is exactly one pdf in the feedback directory, we do not need to
    # create a zip archive.
    if len(feedback_files) == 1 and feedback_files[0].suffix == ".pdf":
        shutil.copy(feedback_files[0], collected_feedback_dir)
        return
    # Otherwise, zip up feedback files.
    feedback_contains_pdf = False
    with ZipFile(
        collected_feedback_dir / collected_feedback_zip_name, "w"
    ) as zip_file:
        for file_to_zip in feedback_files:
            if file_to_zip.suffix == ".pdf":
                feedback_contains_pdf = True
            zip_file.write(
                file_to_zip, arcname=file_to_zip.relative_to(feedback_dir)
            )
    if not feedback_contains_pdf:
        logging.warning(
            f"The feedback for {submission.root_dir.name} contains no PDF file!"
        )


def delete_collected_feedback_directories(sheet: sheets.Sheet) -> None:
    """
    Removes existing collected feedback directories. Does not care about
    non-existing ones.
    """
    for submission in sheet.get_relevant_submissions():
        collected_feedback_dir = submission.get_collected_feedback_dir()
        shutil.rmtree(collected_feedback_dir, ignore_errors=True)


def create_collected_feedback_directories(sheet: sheets.Sheet) -> None:
    """
    Create an empty directory in each relevant team directory. The collected
    feedback will be saved to these directories.
    """
    for submission in sheet.get_relevant_submissions():
        collected_feedback_dir = submission.get_collected_feedback_dir()
        assert not collected_feedback_dir.is_dir() or not any(
            collected_feedback_dir.iterdir()
        )
        collected_feedback_dir.mkdir(exist_ok=True)


def export_xopp_files(sheet: sheets.Sheet) -> None:
    """
    Exports all xopp feedback files.
    """
    logging.info("Exporting .xopp files...")
    for submission in sheet.get_relevant_submissions():
        feedback_dir = submission.get_feedback_dir()
        xopp_files = [
            file for file in feedback_dir.rglob("*") if file.suffix == ".xopp"
        ]
        for xopp_file in xopp_files:
            dest = xopp_file.with_suffix(".pdf")
            subprocess.run(["xournalpp", "-p", dest, xopp_file])
    logging.info("Done exporting .xopp files.")


def print_marks(_the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Prints the marks so that they can be easily copy-pasted to the file where
    marks are collected.
    """
    # Read marks file.
    # Don't check whether the marks file exists because `validate_marks_json()`
    # would have already complained.
    with open(sheet.get_marks_file_path(_the_config), "r", encoding="utf-8") as marks_file:
        marks = json.load(marks_file)

    # Print marks.
    logging.info("Start of copy-paste marks...")
    # We want all teams printed, not just the marked ones.
    for team_to_print in _the_config.teams:
        for submission in sheet.get_all_team_submission_info():
            if submission.team == team_to_print:
                key = submission.team.get_team_key()
                for student in submission.team.members:
                    full_name = f"{student.first_name} {student.last_name}"
                    output_str = f"{full_name:>35};"
                    if _the_config.points_per == "exercise":
                        # The value `marks` assigned to the team_dir key is a
                        # dict with (exercise name, mark) pairs.
                        team_marks = marks.get(key, {"null": ""})
                        _, exercise_marks = zip(*team_marks.items())
                        for mark in exercise_marks:
                            output_str += f"{mark:>3};"
                    else:
                        sheet_mark = marks.get(key, "")
                        output_str += f"{sheet_mark:>3}"
                    print(output_str)
    logging.info("End of copy-paste marks.")


def create_individual_marks_file(_the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Write a json file to add the marks per student.
    """
    with open(sheet.get_marks_file_path(_the_config), "r", encoding="utf-8") as marks_file:
        team_marks = json.load(marks_file)
    student_marks = {}
    for submission in sheet.get_relevant_submissions():
        team_key = submission.team.get_team_key()
        for student in submission.team.members:
            student_key = student.email.lower()
            student_marks.update({student_key: team_marks.get(team_key)})
    file_content = {
        "tutor_name": _the_config.tutor_name,
        "adam_sheet_name": sheet.name,
        "marks": student_marks
    }
    if _the_config.points_per == "exercise" and _the_config.marking_mode == "exercise":
        file_content["exercises"] = sheet.exercises
    with open(sheet.get_individual_marks_file_path(_the_config), "w", encoding="utf-8") as file:
        json.dump(file_content, file, indent=4, ensure_ascii=False)


def create_share_archive(overwrite: Optional[bool], sheet: sheets.Sheet) -> None:
    """
    In case the marking mode is exercise, the final feedback the teams get is
    made up of multiple sets of PDFs (and potentially other files) made by
    multiple tutors. This function stores all feedback made by a single tutor in
    a ZIP file which she/he can then share with the tutor that will send the
    combined feedback. Refer to the `combine` subcommand to see how to process
    the resulting "share archives".
    """
    # This function in only used when the correction mode is 'exercise'.
    # Consequently, exercises must be provided when running 'init', which should
    # be written to sheet.json and then read in when initializing sheets.Sheet.
    assert sheet.exercises
    # Build share archive file name.
    share_archive_file = sheet.get_share_archive_file_path()
    if share_archive_file.is_file():
        # If the user has already chosen to overwrite when considering feedback
        # zips, then overwrite here too. Otherwise, ask here.
        # We should not be here if the user chose 'No' (making overwrite False)
        # before, so we catch this case.
        assert overwrite
        if overwrite is None:
            overwrite = utils.query_yes_no(
                (
                    "There already exists a share archive. Do you want to"
                    " overwrite it?"
                ),
                default=False,
            )
        if overwrite:
            share_archive_file.unlink(missing_ok=True)
        else:
            logging.critical("Aborting 'combine' without overwriting existing share archive.")
    # Take all feedback.zip files and add them to the share archive. The file
    # structure should be similar to the following. In particular, collected
    # feedback that consists of only a single pdf should be zipped to achieve
    # the structure below, whereas collected feedback that is already an archive
    # simply needs to be written under a new name.
    # share_archive_sample_sheet_ex1_ex2.zip
    # ├── 12345_Muster-Meier-Mueller.zip
    # │   └── feedback_tutor1_ex1.pdf
    # ├── 12345_Muster-Meier-Mueller.zip
    # │   ├── feedback_tutor1_ex1.pdf
    # │   └── feedback_tutor1_ex1_code_submission.cc
    # └── ...
    with ZipFile(share_archive_file, "w") as zip_file:
        # The relevant team directories should always be *all* team directories
        # here, because we only need share archives for the 'exercise' marking
        # mode.
        for submission in sheet.get_relevant_submissions():
            collected_feedback_file = submission.get_collected_feedback_path()
            sub_zip_name = f"{submission.root_dir.name}.zip"
            if collected_feedback_file.suffix == ".pdf":
                # Create a temporary zip file in the collected feedback
                # directory and add the single pdf.
                temp_zip_file = (
                    submission.get_collected_feedback_dir() / "temp_zip.zip"
                )
                with ZipFile(temp_zip_file, "w") as temp_zip:
                    temp_zip.write(
                        collected_feedback_file,
                        arcname=collected_feedback_file.name,
                    )
                # Add the temporary zip file to the share archive.
                zip_file.write(temp_zip_file, arcname=sub_zip_name)
                # Remove the temporary zip file.
                temp_zip_file.unlink()
            elif collected_feedback_file.suffix == ".zip":
                zip_file.write(collected_feedback_file, arcname=sub_zip_name)
            else:
                logging.critical(
                    "Collected feedback must be either a single pdf file or a"
                    " single zip archive."
                )


def collect(_the_config: config.Config, args) -> None:
    """
    After marking is done, add feedback files to archives and print marks to be
    copy-pasted to shared point spreadsheet.
    """
    # Prepare.
    sheet = sheets.Sheet(args.sheet_root_dir)
    # Collect feedback.

    # Check if there is a collected feedback directory with files inside
    # already.
    collected_feedback_exists = any(
        (submission.get_collected_feedback_dir()).is_dir()
        and any((submission.get_collected_feedback_dir()).iterdir())
        for submission in sheet.get_relevant_submissions()
    )
    # Ask the user whether collected feedback should be overwritten in case it
    # exists already.
    overwrite = None
    if collected_feedback_exists:
        overwrite = utils.query_yes_no(
            (
                "There already exists collected feedback. Do you want to"
                " overwrite it?"
            ),
            default=False,
        )
        if overwrite:
            delete_collected_feedback_directories(sheet)
        else:
            logging.critical("Aborting 'collect' without overwriting existing collected feedback.")
    if _the_config.xopp:
        export_xopp_files(sheet)
    create_collected_feedback_directories(sheet)
    for submission in sheet.get_relevant_submissions():
        collect_feedback_files(submission, _the_config, sheet)
    if _the_config.marking_mode == "exercise":
        create_share_archive(overwrite, sheet)
    if _the_config.use_marks_file:
        validate_marks_json(_the_config, sheet)
        print_marks(_the_config, sheet)
        create_individual_marks_file(_the_config, sheet)



import logging
from typing import Optional

from .students import *


class Team:
    def __init__(self, members: list[Student], adam_id: Optional[str] = None):
        self.members = members
        self.adam_id = adam_id

    def __eq__(self, other) -> bool:
        return self.members == other.members

    def get_first_names(self) -> list[str]:
        """
        Get a list of the first names of all team members.
        """
        return [member.first_name for member in self.members]

    def get_emails(self) -> list[str]:
        """
        Get a list of the emails of all team members.
        """
        return [member.email for member in self.members]

    def get_team_key(self) -> str:
        """
        Create a string representation of the team with the ADAM ID:
        team_id_LastName1-LastName2
        """
        if not self.adam_id:
            logging.critical("Internal error for developers: get_team_key() "
                             "cannot be used when the team's adam_id is None.")
        return self.adam_id + "_" + self.last_names_to_string()

    def last_names_to_string(self) -> str:
        """
        Concatenate the last names of students to get a pretty-ish string
        representation of a team.
        """
        return "_".join(
            sorted(
                [member.last_name.replace(" ", "-") for member in self.members]
            )
        )

    def to_tuples(self) -> list[tuple[str, str, str]]:
        """
        Get a tuples of strings representation of a team.
        """
        return [member.to_tuple() for member in self.members]


def create_email_to_name_dict(teams: list[Team]) -> dict[str, tuple[str, str]]:
    """
    Maps the students' email addresses to their first and last name.
    """
    email_to_name = {}
    for team in teams:
        for member in team.members:
            email_to_name[member.email] = (member.first_name, member.last_name)
    return email_to_name



import logging
import pathlib
import shutil
from zipfile import ZipFile

from .. import config, sheets, utils


def combine(_the_config: config.Config, args) -> None:
    """
    Combine multiple share archives so that in the end we have one zip archive
    per team containing all feedback for that team.
    """
    # Prepare.
    sheet = sheets.Sheet(args.sheet_root_dir)

    share_archive_files = sheet.get_share_archive_files()
    instructions = (
        "Run `collect` to generate the share archive for your own feedback and"
        " save the share archives you received from the other tutors under"
        f" {sheet.root_dir}."
    )
    if len(list(share_archive_files)) == 0:
        logging.critical(
            f"No share archives exist in {sheet.root_dir}. " + instructions
        )
    if len(list(share_archive_files)) == 1:
        logging.warning(
            "Only a single share archive is being combined. " + instructions
        )

    # Create directory to store combined feedback in.
    combined_dir = sheet.get_combined_feedback_path()
    if combined_dir.exists() and combined_dir.is_dir():
        overwrite = utils.query_yes_no(
            (
                f"The directory {combined_dir} exists already. Do you want to"
                " overwrite it?"
            ),
            default=False,
        )
        if overwrite:
            shutil.rmtree(combined_dir)
        else:
            logging.info(
                f"Could not write to '{combined_dir}'. Aborting combine"
                " command."
            )
            return

    combined_dir.mkdir()

    # Structure at this point:
    # <sheet_root_dir>
    # └── feedback_combined

    # Create subdirectories for teams.
    for submission in sheet.get_relevant_submissions():
        combined_team_dir = combined_dir / submission.root_dir.name
        combined_team_dir.mkdir()

    # Structure at this point:
    # <sheet_root_dir>
    # └── feedback_combined
    #     ├── 12345_Muster-Meier-Mueller
    #     .

    teams_all = [submission.root_dir.name for submission in sheet.get_relevant_submissions()]
    # Extract feedback files from share archives into their respective team
    # directories in the combined directory.
    for share_archive_file in sheet.get_share_archive_files():
        with ZipFile(share_archive_file, mode="r") as share_archive:
            # Check if this share archive is missing team archives for any team.
            teams_present = [
                pathlib.Path(team).stem for team in share_archive.namelist()
            ]
            teams_not_present = list(set(teams_all) - set(teams_present))
            for team_not_present in teams_not_present:
                logging.warning(
                    f"The shared archive {share_archive_file} contains no"
                    f" feedback for team {team_not_present}."
                )
            for team in teams_present:
                # Extract team_archive from share_archive.
                team_archive_file = share_archive.extract(
                    team + ".zip", combined_dir / team
                )
                with ZipFile(team_archive_file, mode="r") as team_archive:
                    team_archive.extractall(path=combined_dir / team)
                pathlib.Path(team_archive_file).unlink()

    """
    I think the step above already accomplishes what this step is supposed to
    accomplish. If no problems arise long-term (today is 2023-10-31), remove
    this block.

    # Structure at this point:
    # <sheet_root_dir>
    # └── feedback_combined
    #     ├── 12345_Muster-Meier-Mueller
    #     .   ├── feedback_exercise_sheet_01_tutor1_ex1.pdf
    #     .   └── feedback_exercise_sheet_01_tutor2_ex2.zip

    # Extract zipped feedback in combined directory.
    for team_dir in combined_dir.iterdir():
        assert team_dir.is_dir()
        for feedback_file in team_dir.iterdir():
            assert feedback_file.is_file()
            # If the feedback is not an archive but a single pdf, move on.
            if feedback_file.suffix != ".zip":
                continue
            # Otherwise, extract feedback from feedback archive.
            with ZipFile(feedback_file, mode="r") as feedback_archive:
                feedback_archive.extractall(path=combined_dir / team)
            # Remove feedback archive.
            feedback_file.unlink()
    """

    # Structure at this point:
    # <sheet_root_dir>
    # └── feedback_combined
    #     ├── 12345_Muster-Meier-Mueller
    #     .   ├── feedback_exercise_sheet_01_tutor1_ex1.pdf
    #     .   ├── feedback_exercise_sheet_01_tutor2_ex2.pdf
    #     .   └── feedback_exercise_sheet_01_tutor2_ex2_code.cc

    # Zip up feedback files.
    for team_dir in combined_dir.iterdir():
        feedback_files = list(team_dir.iterdir())
        combined_team_archive = team_dir / (
            sheet.get_combined_feedback_file_name() + ".zip"
        )
        with ZipFile(combined_team_archive, mode="w") as combined_zip:
            for feedback_file in feedback_files:
                combined_zip.write(feedback_file, arcname=feedback_file.name)
                feedback_file.unlink()

    # Structure at this point:
    # <sheet_root_dir>
    # └── feedback_combined
    #     ├── 12345_Muster-Meier-Mueller
    #     .   └── feedback_exercise_sheet_01.zip



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This script uses the following, potentially ambiguous, terminology:
team     - set of students that submit together, as defined on ADAM
class    - set of teams that are in the same exercise class
relevant - a team whose submission has to be marked by the tutor running the
           script is considered 'relevant'
to mark  - grading/correcting a sheet; giving feedback
marks    - points awarded for a sheet or exercise
"""
import logging
import os
import sys

from . import config, parsers, utils


# Might be necessary to make colored output work on Windows.
os.system("")


def main():
    utils.configure_logging()

    parser = parsers.add_parsers()
    args = parser.parse_args()

    if args.sub_command == "help":
        parser.print_help()
        sys.exit(0)

    # Process config files
    _the_config = config.Config([args.config_shared, args.config_individual])

    # Execute subcommand. This calls the function set as default in the parser.
    # For example, `args.func` is set to `init` if the subcommand is "init".
    logging.info(f"Running command '{args.sub_command}'...")
    args.func(_the_config, args)
    logging.info(f"Command '{args.sub_command}' terminated successfully. 🎉")


if __name__ == "__main__":
    main()



# One-time setup
#    sudo apt install python3-build twine
# or in a venv:
#    pip install build twine

# Publishing a new version:
#
# 1. Update the version tag in this file.
# 2. Remove the `dist/` and the `krummstab.egg-info` directories
# 3. Run the following steps (needs `pip install build twine`):
#
#     $ python3 -m build
#     $ python3 -m twine upload dist/*
#
# 4. Enter the API token

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="krummstab",
    version="0.10",
    description="Efficiently give feedback on ADAM submissions at University of Basel",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Remo Christen",
    author_email="remo.christen@unibas.ch",
    url="https://github.com/aibasel-teaching/krummstab",
    license="GPL3+",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Education",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Topic :: Education",
    ],
    packages=find_packages(),
    python_requires=">=3.10",
    install_requires=[
        'pypdf==3.1.0',
        'jsonschema==4.23.0',
        'xlsxwriter==3.2.0'
    ],
    entry_points={
        'console_scripts': [
            'krummstab = krummstab:main',
        ]
    },
    include_package_data=True,
    package_data={
        "krummstab": ["schemas/config-schema.json", "schemas/submission-info-schema.json"],
    },
)



import json
import logging
import os
import pathlib
import shutil
import tempfile
import textwrap
from typing import Union
from zipfile import ZipFile

from .. import config, sheets, submissions, utils
from ..teams import *


def extract_adam_zip(args) -> tuple[pathlib.Path, str]:
    """
    Unzips the given ADAM zip file and renames the directory to *target* if one
    is given. This is done stupidly right now, it would be better to extract to
    a temporary folder and then move once to the right location.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        # Extract ADAM zip contents to a temporary directory (or move in case it
        # was extracted automatically, this happened on an Apple system).
        if args.adam_zip_path.is_file():
            # Unzip to the directory within the zip file.
            # Should be the name of the exercise sheet,
            # for example "Exercise Sheet 2".
            with ZipFile(args.adam_zip_path, mode="r") as zip_file:
                zip_content = zip_file.namelist()
                sheet_root_dir = pathlib.Path(temp_dir) / zip_content[0]
                utils.filtered_extract(zip_file, pathlib.Path(temp_dir))
        else:
            # Assume the directory is an extracted ADAM zip.
            unzipped_path = pathlib.Path(args.adam_zip_path)
            unzipped_destination_path = (
                pathlib.Path(temp_dir) / unzipped_path.name
            )
            sheet_root_dir = pathlib.Path(
                shutil.copytree(unzipped_path, unzipped_destination_path)
            )
        # Store ADAM exercise sheet name.
        adam_sheet_name = sheet_root_dir.name
        destination = pathlib.Path(
            args.target if args.target else adam_sheet_name
        )
        if destination.exists():
            logging.critical(
                f"Extraction failed because the path '{destination}' exists"
                " already!"
            )
        sheet_root_dir = shutil.move(sheet_root_dir, destination)
    # Flatten intermediate directory.
    sub_dirs = list(sheet_root_dir.glob("*"))
    sub_dirs = [item for item in sub_dirs if item.suffix != '.xlsx']
    if len(sub_dirs) != 1:
        logging.critical(
            "The ADAM zip file contains an unexpected number of"
            f" directories/files ({len(sub_dirs)}), expected exactly 1"
            " subdirectory named either 'Abgaben' or 'Submissions'."
        )
    if sub_dirs[0].name not in ["Abgaben", "Submissions"]:
        logging.warning(
            "It looks like the format of the zip file created by ADAM has"
            " changed. This may require adaptions to this script, but I will"
            " try anyway."
        )
    utils.move_content_and_delete(sub_dirs[0], sheet_root_dir)
    return sheet_root_dir, adam_sheet_name


def mark_irrelevant_team_dirs(_the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Indicate which team directories do not have to be marked by adding the
    `DO_NOT_MARK_PREFIX` to their directory name.
    """
    for submission in sheet.get_all_team_submission_info():
        if not submission.relevant:
            shutil.move(
                submission.root_dir,
                submission.root_dir.with_name(sheets.DO_NOT_MARK_PREFIX +
                                              submission.root_dir.name)
            )


def rename_team_dirs(sheet: sheets.Sheet) -> None:
    """
    The team directories are renamed to: team_id_LastName1-LastName2
    The team ID can be helpful to identify a team on the ADAM web interface.
    """
    for submission in sheet.get_all_team_submission_info():
        team_key = submission.team.get_team_key()
        team_dir = pathlib.Path(
            shutil.move(submission.root_dir, submission.root_dir.with_name(team_key))
        )


def flatten_team_dirs(sheet: sheets.Sheet) -> None:
    """
    There can be multiple directories within a "Team 00000" directory. This
    happens when multiple members of the team upload solutions. Sometimes, only
    one directory contains submitted files, in this case we remove the empty
    ones silently. In case multiple submissions exist, we put the files within
    them next to each other and print a warning.
    """
    for submission in sheet.get_all_team_submission_info():
        # Remove empty subdirectories.
        for team_submission_dir in submission.root_dir.iterdir():
            if team_submission_dir.is_dir() and len(list(team_submission_dir.iterdir())) == 0:
                team_submission_dir.rmdir()
        # Store the list of team submission directories in variable, because the
        # generator may include subdirectories of team submission directories
        # that have already been flattened.
        team_submission_dirs = [path for path in submission.root_dir.iterdir()
                                if not path.name == submissions.SUBMISSION_INFO_FILE_NAME]
        if len(team_submission_dirs) > 1:
            logging.warning(
                f"There are multiple submissions for group '{submission.root_dir.name}'!"
            )
        if len(team_submission_dirs) < 1:
            logging.warning(
                f"The submission of group '{submission.root_dir.name}' is empty!"
            )
        for team_submission_dir in team_submission_dirs:
            if team_submission_dir.is_dir():
                utils.move_content_and_delete(team_submission_dir,
                                              submission.root_dir)


def unzip_internal_zips(sheet: sheets.Sheet) -> None:
    """
    If multiple files are uploaded to ADAM, the submission becomes a single zip
    file. Here we extract this zip. I'm not sure if nested zip files are also
    extracted. Additionally, we flatten the directory as long as a level only
    consists of a single directory.
    """
    for submission in sheet.get_all_team_submission_info():
        for zip_file in submission.root_dir.glob("**/*.zip"):
            with ZipFile(zip_file, mode="r") as zf:
                utils.filtered_extract(zf, zip_file.parent)
            os.remove(zip_file)
        sub_dirs = [path for path in submission.root_dir.iterdir()
                    if not path.name == submissions.SUBMISSION_INFO_FILE_NAME]
        while len(sub_dirs) == 1 and sub_dirs[0].is_dir():
            utils.move_content_and_delete(sub_dirs[0], submission.root_dir)
            sub_dirs = [path for path in submission.root_dir.iterdir()
                        if not path.name == submissions.SUBMISSION_INFO_FILE_NAME]


def create_marks_file(_the_config: config.Config, sheet: sheets.Sheet,
                      args) -> None:
    """
    Write a json file to add the marks for all relevant teams and exercises.
    """
    exercise_dict: Union[str, dict[str, str]] = ""
    if _the_config.points_per == "exercise":
        if _the_config.marking_mode == "static":
            exercise_dict = {
                f"exercise_{i}": "" for i in range(1, args.num_exercises + 1)
            }
        elif _the_config.marking_mode == "exercise":
            exercise_dict = {f"exercise_{i}": "" for i in sheet.exercises}
    else:
        exercise_dict = ""

    marks_dict = {}
    for submission in sorted(sheet.get_relevant_submissions()):
        team_key = submission.team.get_team_key()
        marks_dict.update({team_key: exercise_dict})

    with open(sheet.get_marks_file_path(_the_config), "w", encoding="utf-8") as marks_json:
        json.dump(marks_dict, marks_json, indent=4, ensure_ascii=False)


def create_feedback_directories(_the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Create a directory for every team that should be corrected by the tutor
    specified in the config. A copy of every non-PDF file is prefixed and placed
    in the feedback folder. The idea is that feedback on code files or similar
    can be added to these copies directly and files without feedback can simply
    be deleted. PDFs are not copied, instead a placeholder file with the correct
    name is created so that it can be overwritten by a real PDF file containing
    the feedback.
    """
    for submission in sheet.get_relevant_submissions():
        feedback_dir = submission.get_feedback_dir()
        feedback_dir.mkdir()

        feedback_file_name = sheet.get_feedback_file_name(_the_config)
        dummy_pdf_name = feedback_file_name + ".pdf.todo"
        pathlib.Path(feedback_dir / dummy_pdf_name).touch(exist_ok=True)

        # Copy non-pdf submission files into feedback directory with added
        # prefix.
        for submission_file in submission.root_dir.glob("*"):
            if submission_file.is_dir() or submission_file.suffix == ".pdf" \
                    or submission_file.name == submissions.SUBMISSION_INFO_FILE_NAME:
                continue
            this_feedback_file_name = (
                feedback_file_name + "_" + submission_file.name
            )
            shutil.copy(submission_file, feedback_dir / this_feedback_file_name)


def generate_xopp_files(sheet: sheets.Sheet) -> None:
    """
    Generate xopp files in the feedback directories that point to the single pdf
    in the submission directory and skip if multiple PDF files exist.
    """
    from pypdf import PdfReader

    def write_to_file(f, string):
        f.write(textwrap.dedent(string))

    logging.info("Generating .xopp files...")
    for submission in sheet.get_relevant_submissions():
        pdf_paths = list(submission.root_dir.glob("*.pdf"))
        if len(pdf_paths) != 1:
            logging.warning(
                f"Skipping .xopp file generation for {submission.root_dir.name}: No or"
                " multiple PDF files."
            )
            continue
        pdf_path = pdf_paths[0]
        feedback_dir = submission.get_feedback_dir()
        todo_paths = list(feedback_dir.glob("*.pdf.todo"))
        assert len(todo_paths) == 1
        todo_path = todo_paths[0]
        pages = PdfReader(pdf_path).pages
        # Strips the ".todo" and replaces ".pdf" by ".xopp".
        xopp_path = todo_path.with_suffix("").with_suffix(".xopp")
        if xopp_path.is_file():
            logging.warning(
                f"Skipping .xopp file generation for {submission.root_dir.name}: xopp file"
                " exists."
            )
            continue
        xopp_file = open(xopp_path, "w", encoding="utf-8")
        for i, page in enumerate(pages, start=1):
            width = page.mediabox.width
            height = page.mediabox.height
            if i == 1:  # Special entry for first page
                write_to_file(
                    xopp_file,
                    f"""\
                    <?xml version="1.0" standalone="no"?>
                    <xournal creator="Xournal++ 1.1.1" fileversion="4">
                    <title>Xournal++ document - see https://github.com/xournalpp/xournalpp</title>
                    <page width="{width}" height="{height}">
                    <background type="pdf" domain="absolute" filename="{pdf_path.resolve()}" pageno="{i}"/>
                    <layer/>
                    </page>""",  # noqa
                )
            else:
                write_to_file(
                    xopp_file,
                    f"""\
                    <page width="{width}" height="{height}">
                    <background type="pdf" pageno="{i}"/>
                    <layer/>
                    </page>""",
                )
        write_to_file(xopp_file, "</xournal>")
        xopp_file.close()
        todo_path.unlink()  # Delete placeholder todo file.
    logging.info("Done generating .xopp files.")


def print_missing_submissions(_the_config: config.Config, sheet: sheets.Sheet) -> None:
    """
    Print all teams that are listed in the config file, but whose submission is
    not present in the zip downloaded from ADAM.
    """
    teams_who_submitted = []
    for submission in sheet.get_all_team_submission_info():
        teams_who_submitted.append(submission.team)
    missing_teams = [
        team for team in _the_config.teams if team not in
        teams_who_submitted
    ]
    if missing_teams:
        logging.warning("There are no submissions for the following team(s):")
        for missing_team in missing_teams:
            print(f"* {missing_team.last_names_to_string()}")


def lookup_teams(_the_config: config.Config,
                 team_dir: pathlib.Path) -> tuple[str, list[Team]]:
    """
    Extracts the team ID from the directory name and searches for teams
    based on the extracted email address from the subdirectory name.
    """
    team_id = team_dir.name.split(" ")[1]
    submission_dir = list(team_dir.iterdir())[0]
    submission_email = submission_dir.name.split("_")[-2]
    teams = [
        team
        for team in _the_config.teams
        if any(student.email == submission_email for student in team.members)
    ]
    return team_id, teams


def validate_team_dirs(_the_config: config.Config,
                       sheet_root_dir: pathlib.Path) -> None:
    """
    Checks whether all students are assigned to a team and checks for
    multiple submissions from teams under different IDs.
    """
    adam_id_to_team = {}
    for team_dir in sheet_root_dir.iterdir():
        if not team_dir.is_dir():
            continue
        team_id, teams = lookup_teams(_the_config, team_dir)
        if len(teams) == 0:
            submission_email = list(team_dir.iterdir())[0].name.split("_")[-2]
            logging.critical(
                f"The student with the email '{submission_email}' is not "
                "assigned to a team. Your config file is likely out of date."
                "\n"
                "Please update the config file such that it reflects the team "
                "assignments of this week correctly and share the updated "
                "config file with your fellow tutors and the teaching "
                "assistant."
            )
        # The case that a student is assigned to multiple teams would already be
        # caught when reading in the config file, so we just assert that this is
        # not the case here.
        assert len(teams) == 1
        # TODO: if team[0] in adam_id_to_team.values(): -> multiple separate
        # submissions
        # Catch the case where multiple members of a team independently submit
        # solutions without forming a team on ADAM and print a warning.
        for existing_id, existing_team in adam_id_to_team.items():
            if existing_team == teams[0]:
                logging.warning(
                    f"There are multiple submissions for team '{teams[0].members}'"
                    f" under separate ADAM IDs ({existing_id} and {team_id})!"
                    " This probably means that multiple members of a team"
                    " submitted solutions without forming a team on ADAM. You"
                    " will have to combine the submissions manually."
                )
        adam_id_to_team.update({team_id: teams[0]})


def create_all_submission_info_files(_the_config: config.Config,
                                     sheet_root_dir: pathlib.Path) -> None:
    """
    Creates the submission info JSON files in all team directories.
    """
    for team_dir in sheet_root_dir.iterdir():
        if team_dir.is_dir():
            team_id, teams = lookup_teams(_the_config, team_dir)
            submissions.create_submission_info_file(
                _the_config, teams[0], team_id, team_dir
            )


def init(_the_config: config.Config, args) -> None:
    """
    Prepares the directory structure holding the submissions.
    """
    # Catch wrong combinations of marking_mode/points_per/-n/-e.
    # Not possible eariler because marking_mode and points_per are given by the
    # config file.
    if _the_config.points_per == "exercise":
        if _the_config.marking_mode == "exercise" and not args.exercises:
            logging.critical(
                "You must provide a list of exercise numbers to be marked with "
                "the '-e' flag, for example '-e 1 3 4'. In case the '-e' flag "
                "is the last option before the ADAM zip path, make sure to "
                "separate exercise numbers from the path using '--'."
            )
        if _the_config.marking_mode == "static" and not args.num_exercises:
            logging.critical(
                "You must provide the number of exercises in the sheet with "
                "the '-n' flag, for example '-n 5'."
            )
    else:  # points per sheet
        if _the_config.marking_mode == "exercise":
            logging.critical(
                "Points must be given per exercise if marking is done per "
                "exercise. Set the value of 'poins_per' to 'exercise' or "
                "change the marking mode."
            )
        if args.num_exercises or args.exercises:
            logging.warning(
                "'points_per' is 'sheet', so the flags '-n' and '-e' are "
                "ignored."
            )

    # Sort the list exercises of flag "-e" to make printing nicer later.
    if args.exercises:
        args.exercises.sort()

    # The adam zip file is expected to have the following structure:
    # ADAM Exercise Sheet Name.zip
    # └── ADAM Exercise Sheet Name
    #     └── Abgaben
    #         ├── Team 12345
    #         .   └── Muster_Hans_hans.muster@unibas.ch_000000
    #         .       └── submission.pdf or submission.zip
    sheet_root_dir, adam_sheet_name = extract_adam_zip(args)

    # Structure at this point:
    # <sheet_root_dir>
    # ├── Team 12345
    # .   └── Muster_Hans_hans.muster@unibas.ch_000000
    # .       └── submission.pdf or submission.zip
    validate_team_dirs(_the_config, sheet_root_dir)
    create_all_submission_info_files(_the_config, sheet_root_dir)
    sheet = sheets.create_sheet_info_file(
        sheet_root_dir, adam_sheet_name, _the_config, args.exercises
    )
    print_missing_submissions(_the_config, sheet)

    # Structure at this point:
    # <sheet_root_dir>
    # ├── Team 12345
    # .   ├── Muster_Hans_hans.muster@unibas.ch_000000
    # .   │   └── submission.pdf or submission.zip
    # .   └── submission.json
    # └── sheet.json
    rename_team_dirs(sheet)

    # Structure at this point:
    # <sheet_root_dir>
    # ├── 12345_Muster-Meier-Mueller
    # .   ├── Muster_Hans_hans.muster@unibas.ch_000000
    # .   │   └── submission.pdf or submission.zip
    # .   └── submission.json
    # └── sheet.json
    flatten_team_dirs(sheet)

    # Structure at this point:
    # <sheet_root_dir>
    # ├── 12345_Muster-Meier-Mueller
    # .   ├── submission.pdf or submission.zip
    # .   └── submission.json
    # └── sheet.json
    unzip_internal_zips(sheet)

    # From here on, we need information about relevant teams.
    mark_irrelevant_team_dirs(_the_config, sheet)

    if _the_config.use_marks_file:
        create_marks_file(_the_config, sheet, args)

    create_feedback_directories(_the_config, sheet)

    # Structure at this point:
    # <sheet_root_dir>
    # ├── 12345_Muster-Meier-Mueller
    # .   ├── feedback
    # .   │   ├── feedback.pdf.todo
    # .   │   └── feedback_copy-of-submitted-file.cc
    # .   ├── submission.pdf or submission files
    # .   └── submission.json
    # ├── sheet.json
    # └── points.json
    if _the_config.xopp:
        generate_xopp_files(sheet)



import logging
import pathlib
import shutil
import sys
import tempfile
from zipfile import ZipFile


# Logging ----------------------------------------------------------------------

def configure_logging(level=logging.INFO):
    class ColoredFormatter(logging.Formatter):
        FORMATS = {
            logging.DEBUG: "\033[0;37m[{levelname}]\033[0m {message}",
            logging.INFO: "\033[0;34m[{levelname}]\033[0m {message}",
            logging.WARNING: "\033[0;33m[{levelname}]\033[0m {message}",
            logging.ERROR: "\033[0;31m[{levelname}]\033[0m {message}",
            logging.CRITICAL: "\033[0;31m[{levelname}]\033[0m {message}",
        }

        def format(self, record):
            formatter = logging.Formatter(
                ColoredFormatter.FORMATS[record.levelno], style="{"
            )
            return formatter.format(record)

    class LevelFilter:
        def __init__(self, min_level, max_level):
            self.min_level = min_level
            self.max_level = max_level

        def filter(self, record):
            return self.min_level <= record.levelno <= self.max_level

    class CustomHandler(logging.StreamHandler):
        def __init__(
            self, stream, min_level=logging.DEBUG, max_level=logging.CRITICAL
        ):
            logging.StreamHandler.__init__(self, stream)
            self.setFormatter(ColoredFormatter())
            self.addFilter(LevelFilter(min_level, max_level))

        def emit(self, record):
            logging.StreamHandler.emit(self, record)
            if record.levelno >= logging.CRITICAL:
                sys.exit("aborting")

    root_logger = logging.getLogger("")
    # Remove old handlers
    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)
    root_logger.addHandler(CustomHandler(sys.stdout, max_level=logging.WARNING))
    root_logger.addHandler(CustomHandler(sys.stderr, min_level=logging.ERROR))
    root_logger.setLevel(level)


# Printing ---------------------------------------------------------------------

def query_yes_no(text: str, default: bool = True) -> bool:
    """
    Ask the user a yes/no question and return answer.
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    options = "[Y/n]" if default else "[y/N]"
    print("\033[0;35m[Query]\033[0m " + text + f" {options}")
    choice = input().lower()
    if choice == "":
        return default
    elif choice in valid:
        return valid[choice]
    else:
        logging.warning(
            f"Invalid choice '{choice}'. Please respond with 'yes' or 'no'."
        )
        return query_yes_no(text, default)


# Miscellaneous ----------------------------------------------------------------

def is_hidden_file(name: str) -> bool:
    """
    Check if a given file name could be a hidden file. In particular a file that
    should not be sent to students as feedback.
    """
    return name.startswith(".") or is_macos_path(name)


def is_macos_path(path: str) -> bool:
    """
    Check if the given path is non-essential file created by MacOS.
    """
    return "__MACOSX" in path or ".DS_Store" in path


def filtered_extract(zip_file: ZipFile, dest: pathlib.Path) -> None:
    """
    Extract all files except for MACOS helper files.
    """
    zip_content = zip_file.namelist()
    for file_path in zip_content:
        if is_macos_path(file_path):
            continue
        zip_file.extract(file_path, dest)


def move_content_and_delete(src: pathlib.Path, dst: pathlib.Path) -> None:
    """
    Move all content of source directory to destination directory.
    This does not complain if the dst directory already exists.
    """
    assert src.is_dir() and dst.is_dir()
    with tempfile.TemporaryDirectory() as temp_dir:
        shutil.copytree(src, temp_dir, dirs_exist_ok=True)
        shutil.rmtree(src)
        shutil.copytree(temp_dir, dst, dirs_exist_ok=True)
